import {
  CategoriesResponse,
  EngagementOpportunitiesParameters,
  EngagementParameters,
  EngagementRecommendationsParameters,
  EngagementResponse,
  EngagementsResponse,
  FilteredEngagementsParameters,
  TrackEngagementViewParameters,
  TrackEngagementViewResponse,
} from "./types.ts";
import { createSchema, createYoga, serve, useResponseCache } from "./deps.ts";
import { fetchCategories, fetchEngagement, fetchEngagementOpportunities, trackEngagementView } from "./voltastics.ts";
import { logger } from "./logging.ts";
import { VolunteeringDB, VolunteeringDBConfig } from "./volunteering_db.ts";
import { GeoAPIClient } from "./geo_api_client.ts";
import { Client } from "https://deno.land/x/postgres@v0.19.3/client.ts";

const typeDefs = `
    type Organizer {
      name: String
      imageUrl: String
    }

    type TrackEngagementViewResponse {
      id: String
    }

    type Engagement {
      id: String
      title: String
      description: String
      url: String
      imageUrl: String
      imageUrlOriginal: String
      source: String
      categories: [String]!
      organizer: Organizer
      location: String
      latitude: Float
      longitude: Float
      
      """
      If this engagement is part of a recommendation query response, matchedTopics contains the slugs of matched topics. The order is defined as follows: 1) topics that were part of the query come first (sorted by matching confidence). 2) other topics the engagement matches (sorted by matching confidence)
      """
      matchedTopics: [String]
      
      """
      If this engagement is part of a recommendation query response, matchedSkills contains the slugs of matched skills. The order is defined as follows: 1) skills that were part of the query come first (sorted by matching confidence). 2) other topics the engagement matches (sorted by matching confidence)
      """
      matchedSkills: [String]
      
      """
      If this engagement is part of a recommendation query response, matchedGeoLocationDistanceInMeters contains the distance (in meters) between the engagement and the location given in the query.
      """
      matchedGeoLocationDistanceInMeters: Float
    }

    type Category {
      name: String!
    }

    type CategoriesResponse {
      data: [Category!]!
    }

    type EngagementsResponse {
      totalResults: Int!
      data: [Engagement]!
    }
    
    scalar GeoJSON

    type Query {
        # uses offset-based pagination as described in https://www.apollographql.com/docs/react/pagination/offset-based
        engagementOpportunities(offset: Int! = 0, limit: Int! = 10, location: GeoJSON, category: String): EngagementsResponse!
        
        """
        Retrieves a list of recommendations based on the request parameters "topics", "skills" and "geoLocationId". "topics" contains slugs of the "Topic" type, "skills" contains slugs of the "Skill" type, "geoLocationId" contains the ID of a location as returned by the "geo_*" queries.
        """
        engagementRecommendations(offset: Int! = 0, limit: Int! = 10, topics: [String!] = [], skills: [String!] = [], geolocationId: String): EngagementsResponse!
        
        engagement(id: String!): Engagement
        
        categories: CategoriesResponse!
        filteredEngagements(offset: Int! = 0, limit: Int! = 10, topics: [String!] = [], skills: [String!] = [], geolocationId: String): EngagementsResponse!
    }

    type Mutation {
        trackEngagementView(id: String!): TrackEngagementViewResponse!
    }
`;

const createResolvers = (
  config: ServerConfig,
  volunteeringDB: VolunteeringDB,
) => ({
  Query: {
    engagementOpportunities: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EngagementOpportunitiesParameters,
    ): Promise<EngagementsResponse> =>
      config.fake ? Promise.resolve({ totalResults: 0, data: [] }) : fetchEngagementOpportunities(config)(parameters),
    engagement: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EngagementParameters,
    ): Promise<EngagementResponse> => config.fake ? Promise.resolve(null) : fetchEngagement(config)(parameters),
    categories: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
    ): Promise<CategoriesResponse> => config.fake ? Promise.resolve({ data: [] }) : fetchCategories(config.voltastics),
    filteredEngagements: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: FilteredEngagementsParameters,
    ): Promise<EngagementsResponse> =>
      config.fake ? Promise.resolve({ totalResults: 0, data: [] }) : volunteeringDB.filterEngagements(parameters),
    engagementRecommendations: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EngagementRecommendationsParameters,
    ): Promise<EngagementsResponse> =>
      config.fake ? Promise.resolve({ totalResults: 0, data: [] }) : volunteeringDB.findRecommendations(parameters),
  },
  Mutation: {
    trackEngagementView: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: TrackEngagementViewParameters,
    ): Promise<TrackEngagementViewResponse> =>
      config.fake ? Promise.resolve({ id: parameters.id }) : trackEngagementView(config.voltastics)(parameters),
  },
});

export const DEFAULT_PORT = 8004;
export const DEFAULT_CACHE_ENABLED = true;
export const DEFAULT_CACHE_TTL_MS_VOLTASTICS = 86_400_000; // 24 hours
export const DEFAULT_CACHE_TTL_MS_DB = 3_600_000; // 1 hour

export const DEFAULT_DB_CONNECTION_ATTEMPTS = 10;

export interface VoltasticsConfig {
  baseUrl: string;
  apiToken: string;
}
export interface ServerConfig {
  port: number; // default: 8004
  cacheEnabled: boolean; // default: true
  cacheTtlMsVoltastics: number; // default: 24 hours
  cacheTtlMsDB: number; // default: 1 hour
  voltastics: VoltasticsConfig;
  imageProxyBaseUrl: string;
  fake: boolean; // For local development. If set, the API returns dummy data
  volunteeringDB: VolunteeringDBConfig;
  geoAPIEndpointUrl: string;
}

export const createGraphQLServer = (config: ServerConfig): GraphQLServer => {
  const plugins = config.cacheEnabled
    ? [
      useResponseCache({
        session: () => null, // global cache, shared by all users
        ttlPerSchemaCoordinate: {
          "Query.engagementOpportunities": config.cacheTtlMsVoltastics,
          "Query.engagement": config.cacheTtlMsVoltastics,
          "Query.categories": config.cacheTtlMsVoltastics,
          "Query.filteredEngagements": config.cacheTtlMsDB,
          "Query.engagementRecommendations": config.cacheTtlMsDB,
        },
      }),
    ]
    : [];
  const client = new Client({
    hostname: config.volunteeringDB.hostname,
    port: config.volunteeringDB.port,
    database: config.volunteeringDB.database,
    user: config.volunteeringDB.username,
    password: config.volunteeringDB.password,
    controls: {
      debug: {
        queries: false,
        notices: false,
        results: false,
        queryInError: true,
      },
    },
    connection: {
      attempts: config.volunteeringDB.connectionAttempts,
    },
  });
  const geoAPIClient = new GeoAPIClient(config.geoAPIEndpointUrl);
  const volunteeringDB = new VolunteeringDB(
    client,
    config.imageProxyBaseUrl,
    geoAPIClient,
  );
  const resolvers = createResolvers(config, volunteeringDB);
  return createYoga({
    schema: createSchema({ resolvers, typeDefs }),
    graphiql: true,
    plugins,
  });
};

// deno-lint-ignore no-explicit-any
export type GraphQLServer = any;

export const startServer = (config: ServerConfig): Promise<void> => {
  const graphQLServer: GraphQLServer = createGraphQLServer(config);
  return serve(graphQLServer.handleRequest, {
    port: config.port,
    onListen({ port, hostname }) {
      logger.info(
        `Server started at http://${hostname === "0.0.0.0" ? "localhost" : hostname}:${port}/graphql`,
      );
      if (config.fake) {
        logger.info(
          `Server is serving fake data due to FAKE env var set to true`,
        );
      }
    },
  });
};
